tree grammar ArcheryTreeWalker;


options {
  language = Java;
  tokenVocab = Archery;
  ASTLabelType = CommonTree;
}

@header {
  	package archery.translator;
  	import java.util.LinkedList;
  	import java.util.List;
  	import archery.spec.*; 
  	import archery.spec.process.*;   	 
  	import archery.mcrl2.sort.*;
  	import archery.mcrl2.data.*; 
}

@members {
	SortManager sm = null;
	PatternManager pm = null;
	DataExpFactory def = null; 	
}

specification[Specification s]
@init{
	sm = s.getSortManager();
	pm = s.getPatternManager();
	def = s.getDataExpFactory();
}
	:	^(SPECIFICATION decls* pattern* variable? )
		{ s.setMainVar($variable.value); }
	; 

decls
	: 	^( SORTS sort+ )
	| 	^( CONS 
			( opDecl {	sm.createConstructor($opDecl.value); } )+ 
		)
	|	^( MAP 
			( opDecl {sm.createMap($opDecl.value); } )+ 
		)
	|	^( EQN 
			( eqnDecl { sm.addEquationDecl($eqnDecl.value); } )+ 
		)
	| 	^( VAREQN 
			{ VarEqnBlock veb = new VarEqnBlock(); }
			^( VAR 	( varDecl { veb.addVar($varDecl.value); } )+ ) 
			^( EQN ( eqnDecl { veb.addEqn($eqnDecl.value); } )+ ) 
			{ sm.addVarEqnBlock(veb); }
		)
	; 

opDecl returns[ IdsDecl value ]
	:	^( OP ids sortExp )
		{	value = sm.createIdsDecl($ids.ids, $sortExp.se ); }
	;

eqnDecl returns[EquationDecl value]
	:	^(EQ_SIMPLE d1=dataExp d2=dataExp)
		{ value = sm.createSimpleEquationDecl($d1.value, $d2.value); }
	|	^(EQ_MAP d1=dataExp d2=dataExp d3=dataExp)	
		{ value = sm.createMapEquationDecl($d1.value, $d2.value, $d3.value); }
	;
	
varDecl returns[ IdsDecl value]
 	:	(ids sortExp)
 		{ value = sm.createIdsDecl($ids.ids, $sortExp.se); }
 	;

sort
	:	^(	STANDARD_SORT ids	)
		{	sm.createStandardSort($ids.ids); 	}		
	|   ^( 	STRUCTURED_SORT ids constructors)
		{	sm.createStructuredSort($ids.ids, $constructors.constructors);	}
	|	^( 	NAMED_SORT_EXPRESSION ids sortExp )
		{	sm.createNamedSortExp($ids.ids, $sortExp.se);	}
	;		 

constructors returns [List<Constructor> constructors]
@init{ constructors = new LinkedList<Constructor>(); }
	: 	^( 	CONSTRUCTORS 
			(	constructor {	constructors.add($constructor.cons);	}
			)+ 
		)		 
	;

constructor returns [Constructor cons]
@init{
	String recognizer = null; 			
}
	:	^( 	constructorIdToken=IDENT projections? 
			(recognizerToken=IDENT 	{	recognizer = recognizerToken.getText();	} )?	
		)
		{	
			String constructorId = constructorIdToken.getText();
			cons= sm.createConstructor(constructorId, $projections.projections, recognizer);	
		}
	;
	
projections	 returns [LinkedList<Projection> projections]
@init{ projections = new LinkedList<Projection>(); }
	:	^( 	PROJECTIONS 
			(	projection
				{	projections.add($projection.projection);	}
			)+ 
		)
	;
	
projection returns	[Projection projection]
@init{	String functionName = null;	}
 	:	^( 	PROJECTION 
 			domain 
 			(	i=IDENT
 				{	functionName = i.getText();	}
 			)? )
 		{	projection = sm.createProjection(functionName, $domain.domain);	}
 	;			

domain returns [Domain domain]
@init{	domain = sm.createDomain();	}
	:	^(	DOMAIN 
			(	sortExp
				{	domain.addSort($sortExp.se);	}
			)+
		)
	;

ids returns [List<String> ids]
@init{ ids = new LinkedList<String>(); }
	: 	^( 	IDENTIFIERS  
			(	i=IDENT	{	ids.add(i.getText());	} 
			)+
		)
	;
	
sortExp returns [SortExp se]
	:	// primitive sort
		^(	PRIMITIVE_SORT primitiveSort) 
		{	$se = $primitiveSort.se; }		
	|	// sort reference	
		^( 	SORT_REFERENCE i=IDENT )
		{	$se = sm.createSortReference(i.getText());	}  							
	|	// List, Set, Bag
		^( COLLECTION_SORT collectionType ise=sortExp )
		{	$se = sm.createSortCollection($collectionType.value, $ise.se);	}
	|	^( PARENTHESIZED_SORT_EXPRESSION ise=sortExp)
		{	$se = sm.createParenthesizedSortExp($ise.se);	}
	| 	^( HIGHER_ORDER_SORT domain ise=sortExp )
		{	$se = sm.createHigherOrderSort($domain.domain, $ise.se);}
	;

collectionType returns [int value]
	:	LIST 	{	$value = SortCollection.LIST;	}
	| 	SET 	{	$value = SortCollection.SET;	}
	| 	BAG 	{	$value = SortCollection.BAG;	}
	;

primitiveSort returns [SortExp se]
	:	BOOL 	{	$se = new PrimitiveSort(PrimitiveSort.BOOL);	}
	| 	POS 	{	$se = new PrimitiveSort(PrimitiveSort.POS);	}
	| 	NAT 	{	$se = new PrimitiveSort(PrimitiveSort.NAT);	}
	| 	INT 	{	$se = new PrimitiveSort(PrimitiveSort.INT);	}
	| 	REAL	{	$se = new PrimitiveSort(PrimitiveSort.REAL);	}
	;
	
pattern returns [Pattern p]
	:	^(	PATTERN i=IDENT { p=pm.createPattern(i.getText()); }
			( ^( 	PATTERN_PARAMETERS 
					(	formalParameter
						{ p.addFormalParameters($formalParameter.fps);}
					)+
			) )?
			(	element { p.addElement($element.value); } )+
		)		
	;
	
formalParameter returns [List<FormalParameter> fps]
	:	^(	FORMAL_PARAMETER ids sortExp )
		{	fps = pm.createFormalParameters($ids.ids, $sortExp.se); }
	;

initialisedFormalParameter returns [List<InitializedFormalParameter> ifps]
	:	^(	INITIALIZED_FORMAL_PARAMETER ids sortExp dataExp )
		{	ifps = pm.createInitialisedFormalParameters($ids.ids, $sortExp.se, $dataExp.value); }
	;

	
element returns[Element value]
	:	^( ELEMENT i=IDENT { value = pm.createElement(i.getText()); } 
			( ^(	ELEMENT_PARAMETERS 
					(	formalParameter
					 	{ value.addFormalParameters($formalParameter.fps);}
					)+
			) )?
			^( 	ACTION_BLOCK 
				(	action
					{ value.addActions($action.as);}				
				)+ 
			)				 	
			elementInterface[value]
			^( 	PROCESS_BLOCK 
				processRegistration[value]*
				mainProcess[value] 
				processTranslation[value]* 
			)			
		)
	;
	
action returns [List<Action> as]
	: 	^( 	ACTIONS ids domain? )
		{	as = pm.createActions($ids.ids, $domain.domain); }
	;
	
mainProcess[Element e] returns [MainProcess value]
	:	^( 	MAIN_PROCESS i=IDENT { value = pm.createMainProcess(i.getText(), e); } 
			( ^( 	PROCESS_PARAMETERS 
					(	ifp=initialisedFormalParameter
						{ value.addInitialisedFormalParameters($ifp.ifps); }
					)+ 
			) )?	
			processExp[e] 
			{ value.setProcessExp($processExp.value); }
		)
	;
	
processTranslation[Element e] returns [SubProcess value]
	:	^( PROCESS i=IDENT { value = (SubProcess) e.getProcess(i.getText()); }
			( ^(	PROCESS_PARAMETERS 
					(	formalParameter
						{ value.addFormalParameters($formalParameter.fps); }
					)+
			) )? 
			processExp[e]
			{ value.setProcessExp($processExp.value); }
		)
	;

processRegistration[Element e] returns [SubProcess value]
	:	^( PROCESS i=IDENT { value = pm.createProcess(i.getText(), e); } 
			( ^(PROCESS_PARAMETERS formalParameter+) )? 
			processExpIgnored
		)
	;
	
processExp[Element e] returns[ProcessExp value]
	:	DELTA	
		{ value = pm.createDelta(); }	
	|	TAU		
		{ value = pm.createTau(); }
	|	^( CALL i=IDENT (^(ACTUAL_PARAMETERS dataExps))? )
		{ value = pm.createCall(i.getText(), $dataExps.values, e); }
	| 	^( IF_THEN_ELSE de=dataExp pe1=processExp[e] pe2=processExp[e]? )
		{ value = pm.createConditional($de.value, $pe1.value, $pe2.value); }
	| 	^( PARENTHESIZED_PROCESS_EXPRESSION pe1=processExp[e] )
		{ value = pm.createParenthesisedProcessExp($pe1.value); }
	|	^( SEQUENTIAL_COMPOSITION pe1=processExp[e] pe2=processExp[e] )
		{ value = pm.createSequentialComposition($pe1.value, $pe2.value); }
	| 	^( ALTERNATIVE_COMPOSITION pe1=processExp[e] pe2=processExp[e] )
		{ value = pm.createAlternativeComposition($pe1.value, $pe2.value); }
	; 	

processExpIgnored
	:	DELTA	
	|	TAU		
	|	^( CALL i=IDENT (^(ACTUAL_PARAMETERS dataExps))? )
	| 	^( IF_THEN_ELSE dataExp processExpIgnored processExpIgnored? )
	| 	^( PARENTHESIZED_PROCESS_EXPRESSION processExpIgnored )
	|	^( SEQUENTIAL_COMPOSITION processExpIgnored processExpIgnored )
	| 	^( ALTERNATIVE_COMPOSITION processExpIgnored processExpIgnored )
	; 	



	
elementInterface[Element e] 
	: 	^( 	INTERFACE_BLOCK 
			(	ports[e]
			)+
		)	
	;
	
ports[Element e] 
	: 	^(IN ids)
		{ e.createInPorts($ids.ids); }
	| 	^(OUT ids)
		{ e.createOutPorts($ids.ids); }
	;
	
dataExp returns [DataExp value]
	:	^( DATA_EXP dt=dataTerm der=dataExpRight[$dt.value] )
		{ value = $der.value; }
	| 	dt=dataTerm
	 	{ value = $dt.value; }
	;
	
dataExpRight [DataExp term] returns [DataExp value]
	:	^( 	SUBSTRACTION 
			dt=dataTerm
			{ value=def.createSubstraction(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	|	^( 	ADDITION  
			dt=dataTerm
			{ value=def.createAddition(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)	 		 		 							 		 		 				
	| 	^( 	MULTIPLICATION 
			dt=dataTerm
			{ value=def.createMultiplication(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	INTEGER_DIVISION 
			dt=dataTerm
			{ value=def.createIntDivision(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	REAL_DIVISION 
			dt=dataTerm
			{ value=def.createRealDivision(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	MODULO  
			dt=dataTerm
			{ value=def.createModulo(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	GREATER   
			dt=dataTerm
			{ value=def.createGreater(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	LESS    
			dt=dataTerm
			{ value=def.createLess(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	GREATER_OR_EQUAL     
			dt=dataTerm
			{ value=def.createGreaterOrEqual(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	LESS_OR_EQUAL      
			dt=dataTerm
			{ value=def.createLessOrEqual(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	EQUALITY      
			dt=dataTerm
			{ value=def.createEquality(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		) 
	| 	^( 	DISEQUALITY       
			dt=dataTerm
			{ value=def.createDisequality(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	CONJUNCTION
			dt=dataTerm
			{ value=def.createConjunction(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	DISJUNCTION 
			dt=dataTerm
			{ value=def.createDisjunction(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	IMPLICATION  
			dt=dataTerm
			{ value=def.createImplication(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	LIST_ELEMENT_AT_POSITION   
			dt=dataTerm
			{ value=def.createListElementAtPosition(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	LIST_CONS    
			dt=dataTerm
			{ value=def.createListCons(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	LIST_SNOC     
			dt=dataTerm
			{ value=def.createListSnoc(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	LIST_CONCATENATION      
			dt=dataTerm
			{ value=def.createListConcatenation(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	ELEMENT_TEST       
			dt=dataTerm
			{ value=def.createElementTest(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )? 
		)
	| 	^( 	FUNCTION_UPDATE       	
			^(FUNCTION_UPDATE_DOMAIN dataExps) 
			^(FUNCTION_UPDATE_IMAGE op2=dataExp)
			{	value=def.createFunctionUpdate(term, $dataExps.values, $op2.value); }
			(	der=dataExpRight[value] { value=$der.value; } )?			
		)
	| 	^( 	WHERE 
			dt=dataTerm       	
			{ value=def.createWhere(term, $dt.value); } 
			(	der=dataExpRight[value] { value=$der.value; } )?			
		)
	;
	
dataTerm returns[DataExp value]
	:	i=IDENT 
		{ 	value = def.createIdent(i.getText()); }		
	|	v=BOOL_VALUE
		{	value = def.createBoolean(v.getText()); }		
	|	v=NUMBER
		{	value = def.createNumber(v.getText()); }		
	|	^( LOGICAL_NEGATION op1=dataExp )
		{	value = def.createLogicalNegation($op1.value); }
	|	^( NUMBER_NEGATION op1=dataExp )
		{	value = def.createNumberNegation($op1.value); }
	|	^( LIST_SIZE op1=dataExp )
		{	value = def.createListSize($op1.value); }
	|	^( PARENTHESIZED_DATA_EXPRESSION op1=dataExp)
	 	{	value=def.createParenthesizedDataExp($op1.value); }			 	
	|	^( FUNCTION_APPLICATION i=IDENT dataExps )
	 	{	value=def.createFunctionApplication(i.getText(), $dataExps.values); }		
	|	^( LIST dataExps )
		{	value=def.createList($dataExps.values); } 
	| 	^( SET dataExps )
		{	value=def.createSet($dataExps.values); } 
	| 	^(	BAG bagElements )
		{	value=def.createBag($bagElements.bes); } 	
	|	^(	SET_COMPREHENSION  i=IDENT se=sortExp op1=dataExp)
		{	value=def.createSetComprehension($i.getText(), $se.se, $op1.value); } 		
	|	^(	FORALL i=IDENT se=sortExp op1=dataExp)
		{	value=def.createForAll($i.getText(), $se.se, $op1.value); } 			
	|	^(	EXISTS i=IDENT se=sortExp op1=dataExp)
		{	value=def.createExists($i.getText(), $se.se, $op1.value); } 			
	|	^(	LAMBDA i=IDENT se=sortExp op1=dataExp)	
		{	value=def.createLambda($i.getText(), $se.se, $op1.value); } 			
	;
	
dataExps returns [List<DataExp> values]
@init { values = new LinkedList<DataExp>(); }
	: 	(	dataExp
			{	values.add($dataExp.value); }
		)*
	;
	
bagElements returns [List<BagElement> bes]
@init{ bes = new LinkedList<BagElement>(); }
	: 	(	bagElement
			{	bes.add($bagElement.be); }
		)*
	;
	
bagElement returns [BagElement be]
	:	^(BAG_ELEMENT op1=dataExp op2=dataExp )
		{	be = def.createBagElement($op1.value, $op2.value); }
	;
	
variable returns[Variable value]
	:	^(VARIABLE i=IDENT t=IDENT instance)
		{ value = pm.createVariable(i.getText(), t.getText(), $instance.value); }
	;

instance returns[Instance value]
	:	pi=patternInstance	{ value = $pi.value; }		
	|	ei=elementInstance	{ value = $ei.value; }
	;
	
actualParameters returns [List<DataExp> values]
	:	^(ACTUAL_PARAMETERS dataExps)
		{ values=$dataExps.values; }	
	;

elementInstance returns[ElementInstance value]
	:	^(	ELEMENT_INSTANCE 
			^(	INSTANCE_HEADER t=IDENT args=actualParameters? )
			{ value = pm.createElementInstance(t.getText(), $args.values); }
		)
	;
	
patternInstance	returns[PatternInstance value]
	:	^(	PATTERN_INSTANCE
	  		^( INSTANCE_HEADER t=IDENT args=actualParameters? )
	  		{ value = pm.createPatternInstance(t.getText(), $args.values); }
	  		^( ARCHITECTURE_BODY 
	  			^( 	INSTANCES 
	  				(	v=variable
	  					{ value.addVariable($v.value); }
	  				)+ 
	  			) 
	  			( ^( ATTACHMENTS attachment[value]+ ) )?
	  			( ^( INTERFACE export[value]+ ) )? 
	  		)	  		
		)	
	;
	
attachment[PatternInstance pi] 
	: 	^( ATTACHMENT i1=IDENT p1=IDENT i2=IDENT p2=IDENT )
		{ pi.createAttachment($i1.getText(), $p1.getText(), $i2.getText(), $p2.getText()); }
	;

export[PatternInstance pi]
	:	^( EXPORT i=IDENT p=IDENT r=IDENT )
		{ pi.createPatternPortInstance($i.getText(), $p.getText(), $r.getText()); }
	;